#include <Archivable.h>
#include <String.h>
#include <Message.h>
#include <Messenger.h>
#include <Node.h>

#include <FindDirectory.h>

#include "ResourceSet.h"
#include "othericons.h"

#include "DockArchivableUtils.h"

#include "InnerPanel.h"
#include "InnerPanelIcons.h"

#include "tracker_private.h"

static struct
{
	const char *classname;
	instantiation_func instantiate;
} kStaticDockInstantiationList[] = {
#define LENTRY( x ) { #x, x::Instantiate }
	LENTRY( TInnerPanel ),
	LENTRY( TRaisingIconPanel ),
	LENTRY( TApplicationPanel ),
	LENTRY( TShortcutPanel ),
	LENTRY( TTrackerIcon ),
	LENTRY( TAppPanelIcon ),
	LENTRY( TDockbertIcon ),
	LENTRY( TClockIcon ),
	LENTRY( TTrashIcon ),
	LENTRY( TWorkspacesIcon ),
	LENTRY( TSeparatorIcon ),
	LENTRY( TShowDesktopIcon )
};

const int kStaticDockInstantiationListSize = 12;

BArchivable *instantiate_dock_object( BMessage *msg )
{
	BString classname;
	if ( msg->FindString( "class", &classname ) == B_OK )
	{
		for ( int i=0; i<kStaticDockInstantiationListSize; i++ )
		{
			if ( classname == kStaticDockInstantiationList[i].classname )
				return kStaticDockInstantiationList[i].instantiate( msg );
		}
	}

	return 0;
}

status_t GetEntryIcon( entry_ref *ref, BBitmap *bitmap, icon_size which )
{
/*	BNode node( ref );
	if ( node.InitCheck() != B_OK )
		return B_ERROR;
	BNodeInfo nfo( &node );
	if ( nfo.InitCheck() != B_OK )
		return B_ERROR;

	if ( nfo.GetIcon( bitmap, which ) == B_OK )
		return B_OK;
*/
	BEntry entry( ref, true );

	if ( entry.IsDirectory() )
	{
		// todo: cache this
		BPath path;
		BEntry entry2;
		if ( find_directory( B_USER_DIRECTORY, &path ) == B_OK )
		{
			entry2 = BEntry( path.Path() );

			if ( entry == entry2 )
			{
				const BBitmap *bmp;
				if ( which == B_MINI_ICON )
					bmp = AppResSet()->FindBitmap( 'BBMP', R_HomeIconSmall  );
				else
					bmp = AppResSet()->FindBitmap( 'BBMP', R_HomeIcon );
				memcpy( bitmap->Bits(), bmp->Bits(), bmp->BitsLength() );
				return B_OK;
			}
		}
		if ( find_directory( B_USER_CONFIG_DIRECTORY, &path ) == B_OK )
		{
			path.Append( "be" );
			entry2 = BEntry( path.Path() );

			if ( entry == entry2 )
			{
				const BBitmap *bmp;
				if ( which == B_MINI_ICON )
					bmp = AppResSet()->FindBitmap( 'BBMP', R_NewBeLogoSmall );
				else
					bmp = AppResSet()->FindBitmap( 'BBMP', R_NewBeLogo );
				memcpy( bitmap->Bits(), bmp->Bits(), bmp->BitsLength() );
				return B_OK;
			}
		}

		entry2 = BEntry("/");
		if ( entry == entry2 )
		{
			const BBitmap *bmp;
			if ( which == B_MINI_ICON )
				bmp = AppResSet()->FindBitmap( 'BBMP', R_DisksIconSmall );
			else
				bmp = AppResSet()->FindBitmap( 'BBMP', R_DisksIcon );
			memcpy( bitmap->Bits(), bmp->Bits(), bmp->BitsLength() );
			return B_OK;
		}
	}

	entry_ref real_ref;
	entry.GetRef( &real_ref );

	return BNodeInfo::GetTrackerIcon( &real_ref, bitmap, which );
}

status_t IndependentLaunch( entry_ref &ref )
{
	BMessage message(B_REFS_RECEIVED);
	message.AddRef( "refs", &ref );
	return BMessenger( kTrackerSignature ).SendMessage( &message );
}
