#include "NiceMenu.h"

#include <Bitmap.h>
#include <Screen.h>
#include <Messenger.h>
#include <Roster.h>

#include <stdio.h>

#define _MNITEM_W	4
#define _MNITEM_H	4

static inline float StringHeight( BString /* string */ )
{
	font_height fontHeight;
	be_plain_font->GetHeight(&fontHeight);
	float fTitleAscent = ceilf(fontHeight.ascent);
	float fTitleDescent = ceilf(fontHeight.descent + fontHeight.leading);
	return fTitleAscent + fTitleDescent;
}

TNiceMenuItem::TNiceMenuItem( const char *label )
	: fInvokable(true), fParent(0), fHeight(0), fEnabled(true), fDraggable(false)
{
	SetLabel( BString( label ) );
}

TNiceMenuItem::TNiceMenuItem( const char *label, float height )
	: fInvokable(true), fParent(0), fHeight(0), fEnabled(true), fDraggable(false)
{
	SetLabel( BString( label ) );

	if ( height > fHeight )
		fHeight = height;
}

TNiceMenuItem::TNiceMenuItem( float height )
	: fLabel( "" ), fInvokable(true), fParent(0), fHeight(height), fStringHeight(0), fEnabled(true), fDraggable(false)
{
	fWidth = _MNITEM_W * 2;
}

TNiceMenuItem::~TNiceMenuItem()
{
}

void TNiceMenuItem::SetLabel( BString label )
{
	fLabel = label;
	be_plain_font->TruncateString( &fLabel, B_TRUNCATE_SMART, 200 );
	fStringHeight = StringHeight( fLabel );
	fHeight = fStringHeight + _MNITEM_H * 2 +1;
	fWidth = _MNITEM_W * 8;
	fWidth += ceilf(be_plain_font->StringWidth(fLabel.String()));
}

void TNiceMenuItem::SetEnabled( bool b )
{
	fEnabled = b;
}

void TNiceMenuItem::DrawContent( BRect rect )
{
	if ( !DrawHighlightFrame() )
		fParent->SetLowColor( fParent->fBackgroundColor );

	if ( fLabel.CountChars() > 0 )
	{
		DrawLabelInRect( rect );
	}
}

void TNiceMenuItem::Invalidate()
{
	fParent->Invalidate( fFrame );
}

bool TNiceMenuItem::IsHighlighted()
{
	return fParent->fHighlightedItem == this;
}

void TNiceMenuItem::Invoke()
{
}

void TNiceMenuItem::AttachedToMenu( TNiceMenuView *parent )
{
	fParent = parent;
}

void GonxFrame( BView *view, BRect fullrect, rgb_color backcolor );

bool TNiceMenuItem::DrawHighlightFrame()
{
	if ( fInvokable && fEnabled && IsHighlighted())
	{
/*		BRect r = fFrame;

		r.left += 2;
		r.top += 2;
		r.right -= 2;
		r.bottom -= 1;

		float pensize = fParent->PenSize();
		fParent->SetPenSize( 2.f );

		fParent->SetDrawingMode( B_OP_ALPHA );

		fParent->SetHighColor( 180, 180, 180, 128 );
		fParent->FillRoundRect( r, 7, 7 );
		fParent->SetHighColor( 0, 0, 0, 128 );
		fParent->StrokeRoundRect( r, 7, 7 );

		fParent->SetDrawingMode( B_OP_COPY );
		fParent->SetPenSize( pensize );

		fParent->SetLowColor( 180, 180, 180 );*/
//		GonxFrame( fParent, fFrame, (rgb_color) { 157, 179, 204, 255 } );
		GonxFrame( fParent, fFrame, (rgb_color) { 210, 210, 193, 255 } );
//		fParent->SetLowColor( 157, 179, 204 );
		fParent->SetLowColor( 210, 210, 193 );

		return true;
	}

	return false;
}

void TNiceMenuItem::DrawLabelInRect( BRect r )
{
	if ( fEnabled )
		fParent->SetHighColor( 0, 0, 0 );
	else
		fParent->SetHighColor( 130, 130, 130 );

	BPoint where;
	where.x = r.left + _MNITEM_W;
	where.y = r.bottom - fStringHeight / 2;

	fParent->DrawString( fLabel.String(), where );
}

TNiceMenuItemSeparator::TNiceMenuItemSeparator()
	: TNiceMenuItem( 4 )
{
	fInvokable = false;
}

void TNiceMenuItemSeparator::DrawContent( BRect r )
{
/*	BRect l( r.left + 5, r.top + 3, r.right - 5, r.top +4 );

	fParent->SetHighColor( 255, 255, 255 );
	fParent->FillRect( l );

	l.left ++;
	l.top --; l.bottom --;

	fParent->SetHighColor( 228, 224, 219 );
	fParent->FillRect( l );

	l.top +=2; l.bottom +=2;

	fParent->SetHighColor( 228, 224, 219 );

	for ( int i=0; i<2; i++ )
	{
		l.top ++; l.bottom ++;
		fParent->FillRect( l );
	}*/

//	BRect l( r.left + 2, r.top + 2, r.right - 2, r.top +3 );

/*	float pensize = fParent->PenSize();
	fParent->SetPenSize( 1.5f );

	fParent->SetHighColor( 80, 80, 80, 255 );
	fParent->StrokeRect( l );

	fParent->SetPenSize( pensize );*/
//	GonxFrame( fParent, l, (rgb_color) { 80, 80, 80, 255 } );

	fParent->SetHighColor( 207, 204, 195 );
	fParent->StrokeLine( BPoint( r.left + 2, r.top + 1), BPoint( r.right - 2, r.top +1 ) );
	fParent->StrokeLine( BPoint( r.left + 2, r.bottom - 1), BPoint( r.right - 2, r.bottom -1 ) );
}

TNiceMenuBitmapItem::TNiceMenuBitmapItem( const char *label, BBitmap *bitmap, bool b )
	: TNiceMenuItem( label ), fBitmap(bitmap), fDeleteBitmapInDestructor(b)
{
	fWidth = _MNITEM_W * 8;
	if ( fBitmap )
		fWidth += fBitmap->Bounds().Width();
	fWidth += ceilf(be_plain_font->StringWidth(fLabel.String()));
}

TNiceMenuBitmapItem::~TNiceMenuBitmapItem()
{
	if ( fDeleteBitmapInDestructor && fBitmap )
		delete fBitmap;
}

void TNiceMenuBitmapItem::DrawContent( BRect rect )
{
	if  (!DrawHighlightFrame())
		fParent->SetLowColor( fParent->BackgroundColor() );

	BPoint where;
	where.x = rect.left + 11;
	if ( fBitmap )
	{
		if ( fBitmap->Bounds().Width() > 16 )
			where.x -= 8;
		where.y = rect.top + ( (rect.Height() - fBitmap->Bounds().Height()) / 2);

//		fParent->SetDrawingMode( B_OP_OVER );
		fParent->SetDrawingMode( B_OP_ALPHA );
		if ( !fEnabled )
		{
			fParent->SetBlendingMode( B_CONSTANT_ALPHA, B_ALPHA_OVERLAY );
			fParent->SetHighColor( 255, 255, 255, 100 );
		}
		else
			fParent->SetBlendingMode( B_PIXEL_ALPHA, B_ALPHA_OVERLAY );
		fParent->DrawBitmapAsync( fBitmap, where );
		fParent->SetDrawingMode( B_OP_COPY );

		DrawLabelInRect( BRect( where.x + fBitmap->Bounds().Width() + _MNITEM_W, rect.top, rect.right - _MNITEM_W, rect.bottom) );
	}
	else
	{
		DrawLabelInRect( BRect( where.x, rect.top, rect.right - _MNITEM_W, rect.bottom) );
	}
}

TLauncherMenuItem::TLauncherMenuItem( const char *sig, const char *label, BBitmap *bitmap, bool b )
	: TNiceMenuBitmapItem( label, bitmap, b ), fSignature( sig )
{
}

void TLauncherMenuItem::Invoke()
{
	be_roster->Launch( fSignature.String() );
}

TNiceMessengerMenuItem::TNiceMessengerMenuItem( BMessenger messenger, uint32 message, const char *label, BBitmap *bitmap, bool b )
	: TNiceMenuBitmapItem( label, bitmap, b ), fMessenger( messenger ), fMessage( message )
{
}

void TNiceMessengerMenuItem::Invoke()
{
	fMessenger.SendMessage( fMessage );
}

TNiceMenuView::TNiceMenuView()
	: BView( BRect( 0, 0, 30, 30 ), "MenuView", B_FOLLOW_ALL, B_WILL_DRAW),
		BList(), fHighlightedItem(0), fUseAlphaBlending(false), fAlphedBack(0)
{
	fMenuHeight = 0;

	fBackgroundColor = (rgb_color) { 239, 235, 231, 225 };
	SetViewColor( B_TRANSPARENT_COLOR );

	fLastMouseOverTransit = 0;

	fDragging = false;
	fDraggedItem = 0;

#ifdef B_BEOS_VERSION_DANO
	SetDoubleBuffering( 1 );
#endif
}

TNiceMenuView::~TNiceMenuView()
{
	int32 count = CountItems();
	for ( int i=0; i<count; i++ )
	{
		TNiceMenuItem *item = (TNiceMenuItem*)ItemAt( i );
		delete item;
	}

	if ( fAlphedBack )
		delete fAlphedBack;
}

void TNiceMenuView::Draw( BRect rect )
{
	if ( fUseAlphaBlending )
	{
		DrawBitmapAsync( fAlphedBack, rect, rect );

		SetDrawingMode( B_OP_ALPHA );
	}
	
/*	SetHighColor( fBackgroundColor );
	FillRoundRect( Bounds(), 10, 10 );
	SetHighColor( 80, 80, 80, 255 );
	float pensize = PenSize();
	SetPenSize( 2.f );
	StrokeRoundRect( Bounds(), 10, 10 );
	SetPenSize( pensize );*/
	rgb_color bc = fBackgroundColor;
	if ( !fUseAlphaBlending )
		bc.alpha = 255;
	GonxFrame( this, Bounds(), bc );

	float ywas = 2;
	int32 count = CountItems();
	for ( int i=0; i<count; i++ )
	{
		TNiceMenuItem *item = (TNiceMenuItem*)ItemAt(i);

		item->DrawContent( BRect( 0, ywas, Bounds().Width(), ywas + item->fHeight ) );
		ywas += item->fHeight +1;
	}
}

void TNiceMenuView::MouseDown( BPoint point )
{
	fDraggedItem = ItemAtPoint( point );
}

void TNiceMenuView::MouseUp( BPoint point )
{
	TNiceMenuItem *item = ItemAtPoint( point );

	if ( item && item->fInvokable && item->fEnabled)
		item->Invoke();

	Window()->Quit();
}

void TNiceMenuView::MouseMoved(BPoint point, uint32 transit, const BMessage * /* message */ )
{
	uint32 buttons;

	GetMouse( &point, &buttons );

	// was getting invalid B_ENTERED_VIEW and B_EXITED_VIEW (dano issue?)
	if ( transit == B_ENTERED_VIEW )
	{
		if ( !Bounds().Contains(point) || fLastMouseOverTransit == transit )
			return;
	}
	else if ( transit == B_EXITED_VIEW )
	{
		if ( Bounds().Contains(point) || fLastMouseOverTransit == transit)
			return;
	}

	fLastMouseOverTransit = transit;

	TNiceMenuItem *wasSelected = fHighlightedItem;

	if ( transit == B_INSIDE_VIEW || transit == B_ENTERED_VIEW)
	{
		fHighlightedItem = ItemAtPoint( point );
	}
	else
	{
		fHighlightedItem = 0;

//		if ( !fDragging )
			Window()->Quit();
	}

	if ( fDraggedItem )
	{
		if ( fDraggedItem->fDraggable )
		{
			BMessage *msg = fDraggedItem->BuildDraggingMessage();
			if ( msg )
			{
				fDragging = true;
				DragMessage( msg, fDraggedItem->fFrame, Window() );
				delete msg;
			}
		}
	}

	if ( wasSelected != fHighlightedItem)
	{
		if ( wasSelected )
			wasSelected->Invalidate();
		if ( fHighlightedItem )
			fHighlightedItem->Invalidate();
	}
}

void TNiceMenuView::MessageReceived( BMessage *msg )
{
	if ( msg->WasDropped() )
	{
		dynamic_cast<TNiceMenu*>(Window())->SomethingDropped(msg);
		Window()->Quit();
	}
	else
		BView::MessageReceived(msg);
}

void TNiceMenuView::AddItem( TNiceMenuItem *item )
{
	if ( item->fWidth > Bounds().Width() )
	{
		Window()->ResizeTo( item->fWidth, Bounds().Height() );
		
		for ( int i=0; i<CountItems(); i++ )
		{
			TNiceMenuItem *inner = static_cast<TNiceMenuItem*>( ItemAt(i) );
			inner->fFrame.right = Bounds().Width() - 2; 
		}
	}

	BList::AddItem( item );

	item->fFrame = BRect( 2, Height(), Bounds().Width() - 2, Height() + item->fHeight );

	fMenuHeight += item->fHeight;

	int _y;
	if ( CountItems() == 1 )
		_y = 2;
	else
		_y = 1;

	item->fFrame.top += _y;
	item->fFrame.bottom += _y;
	fMenuHeight += _y;

	item->AttachedToMenu( this );
}

void TNiceMenuView::PrepareAlphaBlending( BRect rect )
{
	if ( fUseAlphaBlending )
	{
		BRect r = rect;
		r.right ++;

		BScreen screen( B_MAIN_SCREEN_ID );

//		hm..
//		if ( r.left < 0 ) r.left = 0;
//		if ( r.right > screen.Frame().Width() ) r.right = screen.Frame().Width();
//		if ( r.top < 0 ) r.top = 0;
//		if ( r.bottom > screen.Frame().Height() ) r.bottom = screen.Frame().Height();

		if ( screen.GetBitmap( &fAlphedBack, false, &r) != B_OK )
		{
			fUseAlphaBlending = false;
		}
	}
}

void TNiceMenuView::SetUseAlphaBlending( bool b )
{
	fUseAlphaBlending = b;
}

TNiceMenu::TNiceMenu( const char *name, BMessenger *target )
	: BWindow( BRect( 0, 0, 30, 30 ), "MenuWindow", B_NO_BORDER_WINDOW_LOOK, B_FLOATING_ALL_WINDOW_FEEL, 0 ),
	fName(name), fTargetMessenger( target )
{
	fMenu = new TNiceMenuView();
	AddChild( fMenu );
}

TNiceMenu::~TNiceMenu()
{
	delete fTargetMessenger;
}

void TNiceMenu::MessageReceived( BMessage *message )
{
	BWindow::MessageReceived(message);
}

bool TNiceMenu::QuitRequested()
{
	BPoint where;
	uint32 buttons;
	fMenu->GetMouse( &where, &buttons );
	return !Bounds().Contains( where );
}

void TNiceMenu::Quit()
{
	BMessage msg( 'nmcl' );
	msg.AddPointer( "menu", this );

	fTargetMessenger->SendMessage( &msg );

	BWindow::Quit();
}

void TNiceMenu::AddItem( TNiceMenuItem *item )
{
	fMenu->AddItem( item );
}

void TNiceMenu::AddSeparatorItem()
{
	AddItem( new TNiceMenuItemSeparator() );
}

void TNiceMenu::SetUseAlphaBlending( bool b )
{
	fMenu->SetUseAlphaBlending( b );
}

void TNiceMenu::GoAt( BRect rect )
{
	ResizeTo( rect.Width(), rect.Height() );
	MoveTo( BPoint( rect.left, rect.top ) );

	fMenu->PrepareAlphaBlending( rect );

	Show();
}
