#ifndef NICEMENU_H
#define NICEMENU_H

#include <View.h>
#include <Window.h>
#include <String.h>
#include <List.h>
#include <Messenger.h>

class BMessage;
class BMessenger;

class TNiceMenuView;

class TNiceMenuItem {
friend class TNiceMenuView;
public:
	TNiceMenuItem( const char * );
	TNiceMenuItem( const char *, float );
	virtual ~TNiceMenuItem();

	void SetLabel( BString );

	void SetEnabled( bool );
	virtual void DrawContent( BRect );
	void Invalidate();
	bool IsHighlighted();

	virtual void Invoke();

	inline float Height() { return fFrame.Height(); }

	virtual BMessage *BuildDraggingMessage() { return 0; }

protected:
	TNiceMenuItem( float height );

	virtual void AttachedToMenu( TNiceMenuView *);
	bool DrawHighlightFrame();
	void DrawLabelInRect( BRect );

	BString fLabel;
	bool fInvokable;
	TNiceMenuView *fParent;
	float fHeight;
	float fWidth;
	float fStringHeight;
	bool fEnabled;
	BRect fFrame;
	bool fDraggable;
};

class TNiceMenuItemSeparator : public TNiceMenuItem {
public:
	TNiceMenuItemSeparator();

	virtual void DrawContent( BRect );
};

class TNiceMenuBitmapItem : public TNiceMenuItem {
public:
	TNiceMenuBitmapItem( const char *label, BBitmap * = 0, bool deleteit = true);
	virtual ~TNiceMenuBitmapItem();

	virtual void DrawContent( BRect );
protected:
	BBitmap *fBitmap;
	bool fDeleteBitmapInDestructor;
};

class TLauncherMenuItem : public TNiceMenuBitmapItem {
public:
	TLauncherMenuItem( const char *sig, const char *label, BBitmap * = 0, bool = true );

	virtual void Invoke();
protected:
	BString fSignature;
};

class TNiceMessengerMenuItem : public TNiceMenuBitmapItem {
public:
	TNiceMessengerMenuItem( BMessenger, uint32, const char *label, BBitmap * = 0, bool = true );

	virtual void Invoke();
protected:
	BMessenger fMessenger;
	uint32 fMessage;
};

class TNiceMenuView : public BView, public BList {
friend class TNiceMenuItem;
public:
	TNiceMenuView();
	virtual ~TNiceMenuView();

	void Draw( BRect );
	virtual void MouseDown( BPoint );
	virtual void MouseUp( BPoint );
	virtual void MouseMoved(BPoint point, uint32 transit, const BMessage *message);
	virtual void MessageReceived( BMessage * );
	virtual void AddItem( TNiceMenuItem * );

	void PrepareAlphaBlending( BRect );

	void SetUseAlphaBlending( bool );
	void SetBackgroundColor( rgb_color );
	rgb_color BackgroundColor() { return fBackgroundColor; }
	inline float Height() { return fMenuHeight; }

	inline TNiceMenuItem *ItemAtPoint( BPoint point )
	{
		float ywas = 2;
		int32 count = CountItems();
		for ( int i=0; i<count; i++ )
		{
			TNiceMenuItem *item = (TNiceMenuItem*)ItemAt(i);

			BRect r( 0, ywas, Bounds().Width(), ywas + item->fHeight );

			if ( r.Contains( point ) )
				return item;

			ywas += item->fHeight +1;
		}
		return 0;
	}

protected:

	TNiceMenuItem *fHighlightedItem;
	rgb_color fBackgroundColor;
	bool fUseAlphaBlending;
	BBitmap *fAlphedBack;
	float fMenuHeight;
	uint32 fLastMouseOverTransit;
	bool fDragging;
	TNiceMenuItem *fDraggedItem;
};

class TNiceMenu : public BWindow {
public:
	TNiceMenu( const char *name = 0, BMessenger * = 0);
	virtual ~TNiceMenu();

	void MessageReceived( BMessage * );

	virtual bool QuitRequested();

	virtual void Quit();

	void AddItem( TNiceMenuItem* );
	void AddSeparatorItem();
	TNiceMenuItem *ItemAt( int );

	void SetUseAlphaBlending( bool );
	void SetBackgroundColor( rgb_color );

	void GoAt( BRect );

	float Width() { return Bounds().Width(); }
	float Height() { return fMenu->Height()+2; }

	void SetMessenger( BMessenger *ms ) { fTargetMessenger = ms; }
	int32 CountItems() { return fMenu->CountItems(); }
	const char *Name() { return fName.String(); }

	virtual void SomethingDropped( BMessage * ) {}

protected:
	BString fName;
	BMessenger *fTargetMessenger;
	TNiceMenuView *fMenu;
};

#endif /* NICEMENU_H */
